from ..algorithms.models import RegressionNetwork
from ..algorithms.models import ClassificationNetwork
from ..algorithms.net import MyNNRegressor, MyNN


__all__ = [
    'ClassificationNetwork',
    'RegressionNetwork',
    'MyNNRegressor',
    'MyNN'
]



